@@ -6,7 +6,7 @@
 ;; Maintainer: Protesilaos Stavrou General Issues <~protesilaos/general-issues@lists.sr.ht>
 ;; URL: https://git.sr.ht/~protesilaos/beframe
 ;; Mailing-List: https://lists.sr.ht/~protesilaos/general-issues
-;; Version: 0.3.0
+;; Version: 1.0.0
 ;; Package-Requires: ((emacs "28.1"))
 
 ;; This file is NOT part of GNU Emacs.
@@ -42,15 +42,24 @@
   :group 'frames)
 
 (defcustom beframe-global-buffers '("*scratch*" "*Messages*" "*Backtrace*")
-  "List of buffer names to include in all frames.
-These buffers are shown in the buffer selection prompts even if
-they have not been used by---and thus associated with---the
-current frame.
+  "List of buffer names (as strings) to include in all frames.
+These buffers are always shown in the `beframe-buffer-menu' or
+buffer selection prompts when `beframe-mode' is enabled.  They do
+not need to be open inside the current frame and thus become
+associated with it (the way other buffers are normally beframed).
 
-Otherwise framed buffers are limited to the frame that uses them."
+When the value is nil, no buffer get this special treatment: they
+all follow the beframing scheme of remaining associated with the
+frame that opened them.
+
+Also see commands such as `beframe-assume-frame-buffers' and
+`beframe-unassume-frame-buffers'.  The full list:
+
+\\{beframe-prefix-map}"
   :group 'beframe
   :package-version '(beframe . "0.1.0")
-  :type '(repeat string))
+  :type '(choice (repeat :tag "List of buffer names as strings" string)
+                 (const :tag "No global buffers" nil)))
 
 (defcustom beframe-create-frame-scratch-buffer t
   "Create a frame-specific scratch buffer for new frames.
@@ -77,7 +86,12 @@ Also see `beframe-create-frame-scratch-buffer'."
 
 When `beframe-mode' is enabled, install advice around each
 function so that every invocation of it is called with
-`other-frame-prefix'."
+`other-frame-prefix'.
+
+Setting the value with `setq' requires a restart of
+`beframe-mode' for changes to take effect.  For changes to apply
+automatically, use `customize-set-variable' or `setopt' (Emacs
+29)."
   :group 'beframe
   :package-version '(beframe . "0.1.0")
   :initialize #'custom-initialize-default
@@ -93,7 +107,12 @@ function so that every invocation of it is called with
 The function accepts one argument, the current frame, as is
 called by the `after-make-frame-functions' hook.
 
-If nil, no renaming is performed."
+If nil, no renaming is performed.
+
+Setting the value with `setq' requires a restart of
+`beframe-mode' for changes to take effect.  For changes to apply
+automatically, use `customize-set-variable' or `setopt' (Emacs
+29)."
   :group 'beframe
   :package-version '(beframe . "0.2.0")
   :initialize #'custom-initialize-default
@@ -113,10 +132,8 @@ If nil, no renaming is performed."
 
 (defun beframe--global-buffers ()
   "Return list of `beframe-global-buffers' buffer objects."
-  (mapcar
-   (lambda (name)
-     (get-buffer name))
-   beframe-global-buffers))
+  (when beframe-global-buffers
+    (mapcar #'get-buffer beframe-global-buffers)))
 
 (cl-defun beframe-buffer-list (&optional frame &key sort)
   "Return list of buffers that are used by the current frame.
@@ -155,17 +172,11 @@ minus all the internal buffers."
 With optional FRAME, do it for the given frame name.  With key
 SORT, apply this sorting function—see `beframe-buffer-list' for
 more information."
-  (mapcar
-   (lambda (buf)
-     (buffer-name buf))
-   (beframe-buffer-list frame :sort sort)))
+  (mapcar #'buffer-name (beframe-buffer-list frame :sort sort)))
 
 (defun beframe--buffer-names-consolidated ()
   "Return list of names of all buffers as strings."
-  (mapcar
-   (lambda (buf)
-     (buffer-name buf))
-   (beframe--buffer-list-consolidated)))
+  (mapcar #'buffer-name (beframe--buffer-list-consolidated)))
 
 (define-obsolete-function-alias
   'beframe--buffer-names
@@ -237,10 +248,18 @@ frame name."
            frame))))
      frames)))
 
-(defun beframe--frame-prompt ()
-  "Prompt to select a frame among the list of frames."
-  (let ((frames (beframe--buffers-with-current)))
-    (completing-read "Select Frame: " frames nil t nil 'frame-name-history (caar frames))))
+(defun beframe--multiple-frames-p ()
+  "Return non-nil if `frame-list' is longer than 1."
+  (> (length (frame-list)) 1))
+
+(defun beframe--frame-prompt (&optional force)
+  "Prompt to select a frame among the list of frames.
+Return user-error if `beframe--multiple-frames-p' is nil.  Skip
+this check if FORCE is non-nil."
+  (if (or force (beframe--multiple-frames-p))
+      (let ((frames (beframe--buffers-with-current)))
+        (completing-read "Select Frame: " frames nil t nil 'frame-name-history (caar frames)))
+    (user-error "Only a single frame is available; aborting")))
 
 (defun beframe--frame-object (frame)
   "Retun frame object of named FRAME.
@@ -335,7 +354,7 @@ Also see the other Beframe commands:
   (display-buffer (beframe-list-buffers-noselect frame :sort sort)))
 
 (defun beframe--assume (buffers)
-  "Unassume BUFFERS from current frame.
+  "Assume BUFFERS into current frame.
 BUFFERS is a list of buffer objects.  If BUFFERS satisfies
 `framep', then the list of buffers is that of the corresponding
 frame object (per `beframe-buffer-list')."
@@ -396,10 +415,7 @@ Also see the other Beframe commands:
 
 (defun beframe--buffers-name-to-objects (buffers)
   "Convert list of named BUFFERS to their corresponding objects."
-  (mapcar
-   (lambda (buf)
-     (get-buffer buf))
-   buffers))
+  (mapcar #'get-buffer buffers))
 
 (defun beframe--buffer-list-prompt-crm (&optional frame)
   "Select one or more buffers in FRAME separated by `crm-separator'.
@@ -541,15 +557,13 @@ Also see the `beframe-prefix-map'."
       (progn
         (setq beframe--read-buffer-function read-buffer-function
               read-buffer-function #'beframe-read-buffer)
-        (add-hook 'after-make-frame-functions #'beframe--frame-predicate)
-        (add-hook 'after-make-frame-functions beframe-rename-function)
-        (add-hook 'after-make-frame-functions #'beframe-create-scratch-buffer)
+        (add-hook 'after-make-frame-functions #'beframe-setup-frame)
+        (add-hook 'context-menu-functions #'beframe-context-menu)
         (beframe--functions-in-frames))
     (setq read-buffer-function beframe--read-buffer-function
           beframe--read-buffer-function nil)
-    (remove-hook 'after-make-frame-functions #'beframe--frame-predicate)
-    (remove-hook 'after-make-frame-functions beframe-rename-function)
-    (remove-hook 'after-make-frame-functions #'beframe-create-scratch-buffer)
+    (remove-hook 'after-make-frame-functions #'beframe-setup-frame)
+    (remove-hook 'context-menu-functions #'beframe-context-menu)
     (beframe--functions-in-frames :disable)))
 
 (defun beframe-create-scratch-buffer (frame)
@@ -583,20 +597,25 @@ Also see the `beframe-prefix-map'."
 ;;     (with-current-buffer buf
 ;;       (rename-buffer (format "*scratch for %s*" name)))))
 
-(defun beframe--infer-frame-name (frame name)
+(defvar project--list) ; from project.el
+
+(defun beframe-infer-frame-name (frame name)
   "Infer a suitable name for FRAME with given NAME.
 See `beframe-rename-frame'."
   (let* ((buffer (car (frame-parameter frame 'buffer-list)))
-         (file-name (when (bufferp buffer)
-                      (buffer-file-name buffer)))
+         (file-name (when (bufferp buffer) (buffer-file-name buffer)))
          (buf-name (buffer-name buffer))
-         (dir (with-current-buffer buffer
-                default-directory)))
+         (dir (with-current-buffer buffer (or (vc-root-dir) default-directory)))
+         (projectp (and (bound-and-true-p project--list)
+                        (listp project--list)
+                        (member (list dir) project--list))))
     (cond
      ((and name (stringp name))
       name)
+     ((and projectp buf-name)
+      (format "%s" (file-name-nondirectory (directory-file-name dir))))
      ((and (not (minibufferp)) file-name)
-      (format "%s  %s" buf-name dir))
+      (format "%s %s" buf-name dir))
      ((not (minibufferp))
       buf-name)
      (t
@@ -624,9 +643,10 @@ With no NAME argument try to infer a name based on the following:
 - Else use the `default-directory'.
 
 Remember that this function doubles as an example for
-`beframe-rename-function': copy it and modify it accordingly."
+`beframe-rename-function': copy it and modify it accordingly
+while also reviewing `beframe-infer-frame-name'."
   (interactive
-   (let ((select-frame (beframe--frame-prompt)))
+   (let ((select-frame (beframe--frame-prompt :force-even-if-one)))
      (list
       (beframe--frame-object select-frame)
       (when current-prefix-arg
@@ -638,18 +658,46 @@ Remember that this function doubles as an example for
   ;;   (beframe--rename-scratch-buffer frame name))
   (modify-frame-parameters
    frame
-   (list (cons 'name (beframe--infer-frame-name frame name)))))
+   (list (cons 'name (beframe-infer-frame-name frame name)))))
+
+(defun beframe-maybe-rename-frame (frame &optional name)
+  "Helper function to determine if `beframe-rename-function' is called.
+FRAME and optional NAME arguments are passed to the
+`beframe-rename-function'."
+  (when beframe-rename-function
+    (funcall beframe-rename-function frame name)))
+
+(defun beframe-setup-frame (frame)
+  "Rename FRAME and create scratch buffer for it, if appropriate.
+Call the functions `beframe-frame-predicate',
+`beframe-do-not-assume-last-selected-buffer',
+`beframe-maybe-rename-frame', `beframe-create-scratch-buffer' in
+this order."
+  (dolist (fn '(beframe-frame-predicate
+                beframe-do-not-assume-last-selected-buffer
+                beframe-maybe-rename-frame
+                beframe-create-scratch-buffer))
+    (funcall fn frame)))
 
 (defun beframe--frame-buffer-p (buf &optional frame)
   "Return non-nil if BUF belongs to the current frame.
 Use optional FRAME to test if BUF belongs to it."
   (memq buf (beframe-buffer-list frame)))
 
-(defun beframe--frame-predicate (&optional frame)
+(define-obsolete-function-alias
+  'beframe--frame-predicate
+  'beframe-frame-predicate
+  "0.4.0")
+
+(defun beframe-frame-predicate (&optional frame)
   "Set FRAME `buffer-predicate' parameter.
 If FRAME is nil, use the current frame."
   (set-frame-parameter frame 'buffer-predicate #'beframe--frame-buffer-p))
 
+(defun beframe-do-not-assume-last-selected-buffer (&rest _)
+  "Unassume the buffer of the most recently used window from the new frame."
+  (beframe--unassume (list (window-buffer (get-mru-window)))))
+
 (defun beframe--with-other-frame (&rest app)
   "Apply APP with `other-frame-prefix'.
 Use this as :around advice to commands that must make a new
@@ -685,5 +733,62 @@ This function can be used as the :sort key of
            (alist-get :visible bufs)
            (alist-get :current bufs))))
 
+;;;; Menu bar and context menu
+
+(defvar beframe--menu-contents
+  '("BEFRAME buffers"
+    ["Switch to beframed buffer" beframe-switch-buffer
+     :help "Switch to a buffer that belongs to the current frame"
+     :enable (beframe--multiple-frames-p)]
+    ["Display beframed buffer menu" beframe-buffer-menu
+     :help "Display a buffer menu consisting of buffers that belong to the current frame"]
+    "---"
+    ["Assume all of a frame's buffers" beframe-assume-frame-buffers
+     :help "Absorb all the buffers of a frame into the current frame buffer list"
+     :enable (beframe--multiple-frames-p)]
+    ["Assume some of a frame's buffers" beframe-assume-frame-buffers-selectively
+     :help "Absorb some buffers of a frame into the current frame buffer list"
+     :enable (beframe--multiple-frames-p)]
+    ["Assume some buffers from all frames" beframe-assume-buffers-selectively-all-frames
+     :help "Absorb some buffers from the global buffer list into the current frame buffer list"
+     :enable (beframe--multiple-frames-p)]
+    ["Assume all buffers outright" beframe-assume-all-buffers-no-prompts
+     :help "Absorb all buffers from the global buffer list into the current frame buffer list"
+     :enable (beframe--multiple-frames-p)]
+    "---"
+    ["Unassume all of a frame's buffers" beframe-unassume-frame-buffers
+     :help "Omit all the buffers of a frame from the current frame buffer list"
+     :enable (beframe--multiple-frames-p)]
+    ["Unassume some of frame's buffers" beframe-unassume-current-frame-buffers-selectively
+     :help "Omit some buffers from the current frame buffer list"
+     :enable (beframe--multiple-frames-p)]
+    ["Unassume all buffers outright" beframe-assume-all-buffers-no-prompts
+     :help "Omit virtually all buffers from the current frame buffer list"
+     :enable (beframe--multiple-frames-p)]
+    "---"
+    ["Toggle Beframe mode" beframe-mode
+     :help "Make all buffer prompts limit candidates per frame (also see `beframe-functions-in-frames')"
+     :style toggle
+     :selected (bound-and-true-p beframe-mode)])
+  "Contents of the Beframe menu.")
+
+(easy-menu-define beframe-global-menu nil
+  "Menu with all Beframe commands, each available in the right context."
+  beframe--menu-contents)
+
+;; Add Beframe menu at the end of global-map after Tools
+(easy-menu-add-item global-map '(menu-bar "Buffers") beframe-global-menu)
+
+(defun beframe-context-menu (menu _click)
+  "Populate MENU with Beframe commands at CLICK."
+  (define-key menu [beframe-separator] menu-bar-separator)
+  (let ((easy-menu (make-sparse-keymap "Beframe")))
+    (easy-menu-define nil easy-menu nil
+      beframe--menu-contents)
+    (dolist (item (reverse (lookup-key easy-menu [menu-bar])))
+      (when (consp item)
+        (define-key menu (vector (car item)) (cdr item)))))
+  menu)
+
 (provide 'beframe)
 ;;; beframe.el ends here
